# -*- coding: utf-8 -*-
from __future__ import division
import logging

from housekeeper.exc import AnalysisNotFinishedError, UnsupportedVersionError
from housekeeper.pipelines.mip.meta import write_meta, build_meta

log = logging.getLogger(__name__)


def prepare_run(segments, force=False):
    """Prepare MIP output."""
    if not force:
        validate(segments['family'])

    outdata_dir = segments['config']['outDataDir']
    qcped_path = segments['family']['PedigreeFileAnalysis']['Path']
    fam_key = segments['config']['familyID']
    customer = segments['family']['InstanceTag'][0]
    case_name = "{}-{}".format(customer, fam_key)
    mip_version = segments['family'].get('MIPVersion', 'v2.x')
    meta_output = build_meta(case_name, segments['family'], qcped_path,
                             version=mip_version)
    write_meta(meta_output, outdata_dir)


def validate(family):
    """Validate analysis."""
    run_status = family['AnalysisRunStatus']
    if run_status != 'Finished':
        log.warn("analysis not finished: %s", run_status)
        raise AnalysisNotFinishedError(run_status)

    mip_version = family.get('MIPVersion')
    if mip_version and mip_version.startswith('v3'):
        log.warn("pipeline too new: %s", mip_version)
        raise UnsupportedVersionError(mip_version)



# -*- coding: utf-8 -*-
from datetime import datetime

from housekeeper.pipelines.general import add


def test_analysis():
    # GIVEN information on a new analysis
    name = 'analysis_1'
    pipeline = 'mip'
    version = 'v3.0.0'
    analyzed_at = datetime.now()
    samples = ['sample_1']
    # WHEN building a new analysis model
    records = add.analysis(name, pipeline, version, analyzed_at,
                           samples=samples)
    # THEN it should return the analysis model with samples
    new_case = records['case']
    new_run = records['run']
    assert new_case.name == name
    assert new_run.pipeline == 'mip'
    assert len(new_run.samples) == 1
    assert new_run.samples[0].name == samples[0]



# -*- coding: utf-8 -*-
from functools import partial

from alchy import Manager
from click.testing import CliRunner
from path import path
import pytest
import yaml

from housekeeper.cli import root
from housekeeper.store import api, Model, Metadata
from housekeeper.pipelines.mip.parse import prepare_inputs


@pytest.fixture
def cli_runner():
    runner = CliRunner()
    return runner


@pytest.fixture
def invoke_cli(cli_runner):
    return partial(cli_runner.invoke, root)


@pytest.fixture
def pedigree():
    return dict(
        checksum='13c366da11437d46e66dfe33cd8cb884d9c1488e',
        path='tests/fixtures/mip/cust003/16074/genomes/16074/16074_pedigree.txt'
    )


@pytest.yield_fixture(scope='function')
def mip_output():
    mip_out = path('tests/fixtures/mip')
    yield mip_out
    mod_qcmetrics = mip_out.joinpath("cust003/16074/analysis/genomes/16074/"
                                     "16074_qcmetrics.mod.yaml")
    mod_qcmetrics.remove_p()
    meta_suffix = "cust003/16074/analysis/genomes/meta.yaml"
    new_meta = mip_out.joinpath(meta_suffix)
    new_meta.remove_p()


@pytest.yield_fixture(scope='function')
def manager():
    config = dict(SQLALCHEMY_DATABASE_URI='sqlite://')
    _manager = Manager(config=config, Model=Model)
    _manager.create_all()
    yield _manager
    _manager.drop_all()


@pytest.yield_fixture(scope='function')
def setup_tmp(tmpdir):
    tmp_path = path(tmpdir)
    tmp_path.joinpath('analyses').makedirs_p()
    db_path = tmp_path.joinpath('store.sqlite3')
    uri = "sqlite:///{}".format(db_path)
    manager = api.manager(uri)
    manager.create_all()
    meta = Metadata(root=tmp_path)
    manager.add_commit(meta)
    data = dict(uri=uri, path=str(db_path), manager=manager)
    yield data
    manager.drop_all()


@pytest.fixture()
def config_path(mip_output):
    suffix = "cust003/16074/genomes/16074/16074_config.yaml"
    _config_path = mip_output.joinpath(suffix)
    return _config_path


@pytest.fixture()
def config_data(config_path):
    with open(config_path, "r") as in_handle:
        _config_data = yaml.load(in_handle)
    return _config_data


@pytest.fixture()
def segments(config_data):
    _segments = prepare_inputs(config_data)
    return _segments



# -*- coding: utf-8 -*-
"""Flask app module."""
import logging
import os

from flask import abort, Flask, render_template, request, redirect, url_for
from flask_alchy import Alchy
from flask_bootstrap import Bootstrap

from housekeeper.store import Model, api

log = logging.getLogger(__name__)

app = Flask(__name__)
SECRET_KEY = os.environ.get('SECRET_KEY') or 'thisIsNotSecret!'
SQLALCHEMY_DATABASE_URI = os.environ['SQLALCHEMY_DATABASE_URI']
if 'mysql' in SQLALCHEMY_DATABASE_URI:  # pragma: no cover
    SQLALCHEMY_POOL_RECYCLE = 3600
SQLALCHEMY_TRACK_MODIFICATIONS = False
TEMPLATES_AUTO_RELOAD = True
app.config.from_object(__name__)

Bootstrap(app)
db = Alchy(app, Model=Model)


@app.route('/')
def index():
    """Just redirect for now."""
    return redirect(url_for('cases'))


@app.route('/cases')
def cases():
    """Overview all loaded cases."""
    page = int(request.args.get('page', '1'))
    qargs = {
        'query_str': request.args.get('query_str'),
        'per_page': 20,
    }
    cases_q = api.cases(query_str=qargs['query_str'])
    cases_page = cases_q.paginate(page, per_page=qargs['per_page'])
    return render_template('cases.html', cases=cases_page, qargs=qargs)


@app.route('/cases/<name>')
def case(name):
    """Show more information about a case."""
    case_obj = api.case(name)
    if case_obj is None:
        return abort(404)
    return render_template('case.html', case=case_obj)


@app.route('/cases/<case_name>/postpone', methods=['POST'])
def case_postpone(case_name):
    """Postpone clean up date for the latest run of a case."""
    run_obj = api.runs(case_name).first()
    if run_obj is None:
        return abort(404)
    api.postpone(run_obj)
    db.commit()
    return redirect(request.referrer)



# -*- coding: utf-8 -*-
import logging

from housekeeper.store import AnalysisRun, Case, Sample
from housekeeper.constants import TIME_TO_CLEANUP

log = logging.getLogger(__name__)


def analysis(name, pipeline, version, analyzed_at, samples=None):
    """Store information about a general analysis run.

    This is the most low level implementation of how to store files from an
    analysis.
    """
    new_case = Case(name=name)
    new_run = AnalysisRun(pipeline=pipeline, pipeline_version=version,
                          analyzed_at=analyzed_at)

    # set the future date for archiving
    new_run.will_cleanup_at = analyzed_at + TIME_TO_CLEANUP

    for sample_id in (samples or []):
        new_run.samples.append(Sample(name=sample_id))
    return {'case': new_case, 'run': new_run}



# -*- coding: utf-8 -*-
import datetime
import logging

from alchy import Manager
from path import path

from .models import AnalysisRun, Asset, Case, Model, Sample
from .utils import get_rundir

log = logging.getLogger(__name__)


def manager(db_uri):
    log.debug('open connection to database: %s', db_uri)
    db = Manager(config=dict(SQLALCHEMY_DATABASE_URI=db_uri), Model=Model,
                 session_options=dict(autoflush=False))
    return db


def assets(case_name=None, sample=None, category=None, run_id=None):
    """Get files from the database."""
    query = Asset.query
    if run_id:
        query = query.filter_by(run_id=run_id)
    elif case_name:
        query = (query.join(Asset.run, AnalysisRun.case)
                      .filter(Case.name == case_name))
    if sample:
        query = query.join(Asset.sample).filter(Sample.name == sample)
    if category:
        query = query.filter(Asset.category == category)
    return query


def case(name):
    """Get a case from the database.

    Args:
        name (str): the unique name of the analysis

    Returns:
        Case or None: the case or None if not found
    """
    # 'name' is a unique column so it will always return 1 or 0 records
    case_obj = Case.query.filter_by(name=name).first()
    return case_obj


def samples():
    """Get samples from the database."""
    sample_query = Sample.query
    return sample_query


def sample(name):
    """Get sample from the database."""
    sample_obj = Sample.query.filter_by(name=name).first()
    return sample_obj


def cases(query_str=None):
    """Get multiple cases from the database."""
    query = Case.query.order_by(Case.created_at.desc())
    if query_str:
        query = query.filter(Case.name.like("%{}%".format(query_str)))
    return query


def runs(case_name=None, run_date=None, since=None):
    """Get runs for a case from the database.

    Args:
        case_name (str): the unique name of the case

    Returns:
        Query: all runs for the case
    """
    if run_date:
        delta = datetime.timedelta(days=1)
        next_day = run_date + delta
        condition = AnalysisRun.analyzed_at.between(run_date, next_day)
        return AnalysisRun.query.filter(condition)
    run_query = AnalysisRun.query.order_by(AnalysisRun.created_at.desc())
    if since:
        run_query = run_query.filter(AnalysisRun.created_at > since)
    if case_name:
        run_query = (run_query.join(AnalysisRun.case)
                              .filter(Case.name == case_name))
    return run_query


def delete(run_obj):
    """Delete an analysis run along with related files on the system.

    Args:
        run_obj (AnalysisRun): the analysis run to delete
    """
    run_dir = get_rundir(run_obj.case.name, run_obj)
    delete_dir(run_dir)

    if len(run_obj.case.runs) == 1:
        run_obj.case.delete()
    else:
        run_obj.delete()


def delete_asset(asset_obj):
    """Delete an asset completely from the database."""
    if asset_obj.is_local:
        path(asset_obj.path).remove_p()
    asset_obj.delete()


def clean_up(run_obj, force=False):
    """Clean up files for an analysis."""
    # check if run is ready
    if not force and not all([run_obj.archived_at, run_obj.delivered_at]):
        # not ready for clean up!
        log.warn("run not ready for clean up")
    else:
        for asset in run_obj.assets:
            if not asset.archive_type:
                asset.delete()
            else:
                asset.is_local = False

        run_dir = get_rundir(run_obj.case.name, run_obj)
        log.info("removing rundir: %s", run_dir)
        delete_dir(run_dir)
        run_obj.cleanedup_at = datetime.datetime.now()


def postpone(run_obj, time=datetime.timedelta(days=30)):
    """Postpone the automatic archival of analysis by X time."""
    run_obj.will_cleanup_at += time


def delete_dir(directory):
    directory_path = path(directory)
    if directory_path.exists():
        log.info("removing files under: %s", directory)
        directory_path.rmtree_p()
    else:
        log.debug("run directory not found: %s", directory)


def add_asset(run_obj, asset_path, category, archive_type=None, sample=None):
    """Link an asset to a run."""
    abs_path = path(asset_path).abspath()
    new_asset = Asset(original_path=abs_path, category=category,
                      archive_type=archive_type)
    new_asset.sample = sample
    return new_asset



# -*- coding: utf-8 -*-
import json

from housekeeper.store import Sample


def test_to_json():
    # GIVEN a basic sample model
    sample_name = 'ADM12'
    new_sample = Sample(name=sample_name)
    # WHEN serializing to JSON
    sample_json = new_sample.to_json()
    # THEN it should return a serialized dict
    data = json.loads(sample_json)
    assert data['name'] == sample_name



# -*- coding: utf-8 -*-
import logging
import pkg_resources

import click
from path import path
import yaml

from housekeeper.store import api
from housekeeper.store.utils import get_rundir
from housekeeper.cli.utils import run_orabort
from housekeeper import exc
from .mip import parse_mip
from .mip2 import parse as parse_mip2
from .miparchive import parse as parse_miparchive
from .general import commit_analysis, check_existing

log = logging.getLogger(__name__)

LOADERS = {'mip': parse_mip, 'mip2': parse_mip2,
           'miparchive': parse_miparchive}


@click.command()
@click.option('-y', '--yes', is_flag=True, help='auto replace old runs')
@click.option('-f', '--force', is_flag=True, help='replace identical runs')
@click.option('-r', '--references', type=click.File('r'))
@click.option('-p', '--pipeline', type=click.Choice(LOADERS.keys()),
              default='mip')
@click.argument('config', type=click.File('r'))
@click.pass_context
def add(context, force, yes, references, pipeline, config):
    """Add analyses from different pipelines."""
    manager = api.manager(context.obj['database'])
    config_data = yaml.load(config)
    if not references:
        default_ref = "pipelines/references/{}.yaml".format(pipeline)
        references = pkg_resources.resource_string("housekeeper", default_ref)
    reference_data = yaml.load(references)
    loader = LOADERS[pipeline]
    try:
        records = loader(config_data, reference_data, force=force)
    except (exc.UnsupportedVersionError, exc.AnalysisNotFinishedError):
        context.abort()
    case_name = records['case'].name
    old_run = check_existing(case_name, records['run'])
    if old_run:
        click.echo("identical run detected: {}".format(case_name))
        if not force:
            context.abort()
        else:
            api.delete(old_run)

    try:
        commit_analysis(manager, **records)
        click.echo("added new analysis: {}".format(case_name))
        sample_ids = ', '.join(sample.name for sample in records['run'].samples)
        click.echo("including samples: {}".format(sample_ids))
    except exc.AnalysisConflictError:
        click.echo("analysis output not removed: {}".format(case_name))


@click.command()
@click.option('-d', '--date', help="date of the particular run")
@click.option('-c', '--category', required=True)
@click.option('-s', '--sample', help='sample id to link assets to')
@click.option('-a', '--archive-type', type=click.Choice(['data', 'result']))
@click.argument('case_name')
@click.argument('asset_path')
@click.pass_context
def extend(context, date, category, sample, archive_type, case_name, asset_path):
    """Add an additional asset to a run."""
    manager = api.manager(context.obj['database'])
    run_obj = run_orabort(context, case_name, date)
    if sample:
        sample_map = {smpl.name: smpl for smpl in run_obj.samples}
        sample_obj = sample_map[sample]
    else:
        sample_obj = None
    new_asset = api.add_asset(run_obj, asset_path, category, archive_type,
                              sample=sample_obj)

    run_root = get_rundir(run_obj.case.name, run_obj)
    filename = new_asset.basename()
    new_path = run_root.joinpath(filename)
    new_asset.path = new_path

    log.debug("link asset: %s -> %s", new_asset.original_path, new_asset.path)
    path(new_asset.original_path).link(new_asset.path)
    run_obj.assets.append(new_asset)
    log.info("add asset: %s", new_asset.path)
    manager.commit()



# -*- coding: utf-8 -*-
import datetime
import logging

import click

from housekeeper.cli.utils import run_orabort
from .utils import get_rundir
from . import api, Asset, AnalysisRun, Sample, Case

log = logging.getLogger(__name__)


@click.command()
@click.option('-c', '--case')
@click.option('-s', '--sample')
@click.option('-i', '--infer-case', is_flag=True)
@click.option('-c', '--category')
@click.pass_context
def get(context, case, sample, infer_case, category):
    """Ask Housekeeper for a file."""
    api.manager(context.obj['database'])
    if infer_case:
        sample_obj = api.sample(sample)
        if sample_obj is None:
            log.warn('sorry, sample not found')
            context.abort()
        case = sample_obj.run.case.name
        sample = None
    assets = api.assets(case_name=case, sample=sample, category=category)
    paths = [asset.path for asset in assets]
    output = ' '.join(paths)
    click.echo(output, nl=False)


@click.command()
@click.option('-d', '--days', default=30)
@click.argument('case_name')
@click.pass_context
def postpone(context, days, case_name):
    """Ask Housekeeper for a file."""
    manager = api.manager(context.obj['database'])
    run_obj = run_orabort(case_name)
    api.postpone(run_obj, time=datetime.timedelta(days=days))
    manager.commit()
    click.echo("analysis will be archived on: {}"
               .format(run_obj.will_cleanup_at))


@click.command()
@click.option('-y', '--yes', is_flag=True, help="skip confirmation")
@click.option('-d', '--date', help="date of the particular run")
@click.argument('case_name')
@click.pass_context
def delete(context, date, yes, case_name):
    """Delete an analysis run and files."""
    manager = api.manager(context.obj['database'])
    run_obj = run_orabort(context, case_name, date)
    run_root = get_rundir(case_name, run_obj)
    click.echo("you are about to delete: {}".format(run_root))
    if yes or click.confirm('Are you sure?'):
        api.delete(run_obj)
        manager.commit()


@click.command()
@click.option('-f', '--force', is_flag=True)
@click.option('-d', '--date', help="date of the particular run")
@click.argument('case_name')
@click.pass_context
def clean(context, force, date, case_name):
    """Clean up files for an analysis."""
    manager = api.manager(context.obj['database'])
    run_obj = run_orabort(context, case_name, date)
    if force or click.confirm('Are you sure?'):
        api.clean_up(run_obj, force=force)
        manager.commit()


@click.command()
@click.option('-l', '--limit', default=20)
@click.option('-s', '--since', help='consider runs since date')
@click.option('-c', '--category', default='bcf-raw')
@click.pass_context
def ls(context, limit, since, category):
    """List files from recently added runs."""
    api.manager(context.obj['database'])
    if category == 'samples':
        query = api.samples().join(Sample.run)
    elif category == 'cases':
        query = AnalysisRun.query
    else:
        query = api.assets(category=category).join(Asset.run)

    query = query.order_by(AnalysisRun.analyzed_at.desc())
    if since:
        date_obj = build_date(since) if since else None
        query = query.filter(AnalysisRun.analyzed_at >= date_obj)
    else:
        query = query.limit(limit)

    if query.first() is None:
        log.warn('sorry, no matching assets found')
        context.abort()

    if category == 'samples':
        # we only consider unqiue samples
        output = set(sample.name for sample in query)
    elif category == 'cases':
        output = set(run.case.name for run in query)
    else:
        # we only consider files from the latest run per case
        cases = set()
        output = []
        for asset in query:
            if asset.run.case_id not in cases:
                output.append(asset.path)
                cases.add(asset.run.case_id)
    click.echo(" ".join(output))


def build_date(date_str):
    """Parse date out of string."""
    return datetime.date(*map(int, date_str.split('-')))



    # -*- coding: utf-8 -*-
from path import path

from housekeeper import initiate


def test_init(invoke_cli, tmpdir):
    # GIVEN a non-existing directory
    root_path = tmpdir.join('keeper')
    db_path = tmpdir.join('store.sqlite3')
    db_uri = path("sqlite:///{}".format(str(db_path)))
    assert not path(root_path).exists()
    # WHEN executing the init command
    result = invoke_cli(['--database', db_uri, 'init', str(root_path)])
    # THEN it should return successfully
    assert path(root_path).exists()
    assert result.exit_code == 0


def test_init_existing(invoke_cli, tmpdir):
    # GIVEN an existing directory
    db_path = tmpdir.join('store.sqlite3')
    db_uri = path("sqlite:///{}".format(str(db_path)))
    assert path(tmpdir).exists()
    assert not db_path.exists()
    # WHEN running the init command
    result = invoke_cli(['--database', db_uri, 'init', str(tmpdir)])
    # THEN it should exit unsuccesfully
    assert result.exit_code != 0
    assert not db_path.exists()



# -*- coding: utf-8 -*-
from housekeeper.store import AnalysisRun


def test_add_mip(invoke_cli, mip_output, setup_tmp):
    # GIVEN some MIP analysis output
    assert AnalysisRun.query.first() is None
    # WHEN adding a MIP analysis on the command line
    config = mip_output.joinpath('cust003/16074/genomes/16074/16074_config.yaml')
    result = invoke_cli(['-d', setup_tmp['uri'], 'add', config])
    # THEN it should add the analysis and files to the database
    assert result.exit_code == 0
    assert AnalysisRun.query.first().case.name == 'cust003-16074'



# -*- coding: utf-8 -*-


class AnalysisConflictError(Exception):
    pass


class MissingFileError(Exception):
    pass


class AnalysisNotFinishedError(Exception):
    pass


class UnsupportedVersionError(Exception):
    pass


class MalformattedPedigreeError(Exception):
    pass



# -*- coding: utf-8 -*-
import logging

from housekeeper.store import api, utils

log = logging.getLogger(__name__)


def run_orabort(context, case_name, date_str=None):
    """Get a run or abort the Click context."""
    run_date = utils.build_date(date_str) if date_str else None
    run_obj = api.runs(case_name, run_date=run_date).first()
    if run_obj is None:
        log.error("no analysis run found for case: %s", case_name)
        context.abort()
    else:
        return run_obj



# -*- coding: utf-8 -*-
from housekeeper.compile import core


def test_checksum(pedigree):
    # GIVEN a file with checksum
    checksum = pedigree['checksum']
    path = pedigree['path']
    # WHEN calculating the checksum
    calculated_checksum = core.checksum(path)
    # THEN it should come up the same as expected
    assert calculated_checksum == checksum



# -*- coding: utf-8 -*-
from collections import namedtuple
from datetime import datetime
import hashlib
import logging

from path import path

from housekeeper.store import api
from housekeeper.store.utils import get_rundir
from .tar import tar_files

BLOCKSIZE = 65536
ArchiveGroup = namedtuple('ArchiveGroup', ['id', 'out', 'checksum'])
log = logging.getLogger(__name__)


def compile_run(run_obj):
    """High level compile function for a run."""
    groups = compress_run(run_obj)
    for group in groups:
        log.info("compressed %s archive: %s", group.id, group.out)
        category = "archive-{}".format(group.id)
        new_asset = api.add_asset(run_obj, group.out, category)
        new_asset.path = group.out
        new_asset.checksum = group.checksum
        run_obj.assets.append(new_asset)
    run_obj.compiled_at = datetime.now()


def compress_run(run_obj):
    """Archive a run into data+results archives."""
    run_dir = path(get_rundir(run_obj.case.name, run=run_obj))
    for group, paths in group_assets(run_obj.assets):
        run_date = run_obj.analyzed_at.date()
        group_out = run_dir.joinpath("{}.{}.tar.gz".format(run_date, group))
        filenames = [path(full_path).basename() for full_path in paths]
        tar_files(group_out, root_dir=run_dir, filenames=filenames)
        sha1 = checksum(group_out)
        yield ArchiveGroup(id=group, out=group_out, checksum=sha1)


def group_assets(assets):
    """Groups asset paths based on archive type."""
    # group files base on type
    groups = ('data', 'result')
    for group in groups:
        paths = [asset.path for asset in assets
                 if asset.archive_type in (group, 'meta')]
        yield group, paths


def checksum(path):
    """Calculcate checksum for a file."""
    hasher = hashlib.sha1()
    with open(path, 'rb') as stream:
        buf = stream.read(BLOCKSIZE)
        while len(buf) > 0:
            hasher.update(buf)
            buf = stream.read(BLOCKSIZE)
    return hasher.hexdigest()



# -*- coding: utf-8 -*-
from __future__ import division
from housekeeper.pipelines.mip import prepare


def test_total_mapped():
    # GIVEN the path to the samtools multiqc summary
    path = ("tests/fixtures/mip/cust003/16074/analysis/genomes/ADM1594A12/"
            "multiqc/multiqc_data/multiqc_samtools.txt")
    # WHEN calculating the overall mapping rate
    with open(path, 'r') as stream:
        mapped_reads = prepare.total_mapped(stream)
    # THEN it should return the correct value
    assert mapped_reads['percentage'] == 891591355 / 920254494


def test_prepare_run(mip_output, segments):
    # GIVEN a MIP analysis without alterations
    mod_qcmetrics = mip_output.joinpath("cust003/16074/analysis/genomes/16074/"
                                        "16074_qcmetrics.mod.yaml")
    meta_suffix = "cust003/16074/analysis/genomes/meta.yaml"
    new_meta = mip_output.joinpath(meta_suffix)
    assert not mod_qcmetrics.exists()
    assert not new_meta.exists()
    # WHEN collecting information/preparing it for housekeeper
    prepare.prepare_run(segments)
    # THEN it should produce the modded qc metrics file
    assert mod_qcmetrics.exists()
    assert new_meta.exists()



# -*- coding: utf-8 -*-
from housekeeper.store import utils


def test_build_date():
    # GIVEN a date string in the iso format
    date_str = '2015-02-19'
    # WHEN converting to a date object
    date_obj = utils.build_date(date_str)
    # THEN it should parse out the values
    assert date_obj.day == 19
    assert date_obj.month == 2
    assert date_obj.year == 2015



# -*- coding: utf-8 -*-
from datetime import datetime
import json

import alchy
from path import path
from sqlalchemy import Column, ForeignKey, orm, types, UniqueConstraint

from housekeeper.constants import PIPELINES, ARCHIVE_TYPES


def json_serial(obj):
    """JSON serializer for objects not serializable by default json code"""
    if isinstance(obj, datetime):
        serial = obj.isoformat()
        return serial
    raise TypeError('Type not serializable')


class JsonModel(alchy.ModelBase):

    def to_json(self, pretty=False):
        """Serialize Model to JSON."""
        kwargs = dict(indent=4, sort_keys=True) if pretty else dict()
        return json.dumps(self.to_dict(), default=json_serial, **kwargs)


Model = alchy.make_declarative_base(Base=JsonModel)


class Metadata(Model):
    """Store information about the system."""

    id = Column(types.Integer, primary_key=True)
    created_at = Column(types.DateTime, default=datetime.now)
    root = Column(types.String(128), nullable=False)

    @property
    def root_path(self):
        return path(self.root)


class Case(Model):
    """Store a permanent reference to a case."""

    id = Column(types.Integer, primary_key=True)
    name = Column(types.String(128), unique=True)
    created_at = Column(types.DateTime, default=datetime.now)

    runs = orm.relationship('AnalysisRun', cascade='all,delete',
                            backref='case', order_by='-AnalysisRun.id')

    @property
    def current(self):
        """Return the current (active) run."""
        return self.runs[0]


class AnalysisRun(Model):
    """Store information about a specific analysis run."""

    __table_args__ = (UniqueConstraint('case_id', 'analyzed_at',
                                       name='_uc_analysis_analyzed_at'),)

    id = Column(types.Integer, primary_key=True)
    created_at = Column(types.DateTime, default=datetime.now)
    pipeline = Column(types.Enum(*PIPELINES))
    pipeline_version = Column(types.String(32))
    analyzed_at = Column(types.DateTime)
    compiled_at = Column(types.DateTime)
    delivered_at = Column(types.DateTime)
    archived_at = Column(types.DateTime)
    cleanedup_at = Column(types.DateTime)
    will_cleanup_at = Column(types.DateTime)

    case_id = Column(types.Integer, ForeignKey('case.id'), nullable=False)
    assets = orm.relationship('Asset', cascade='all,delete', backref='run')
    samples = orm.relationship('Sample', cascade='all,delete', backref='run')

    @property
    def cleanup_in(self):
        """Return number of days until archive happens."""
        time_diff = self.will_cleanup_at - datetime.today()
        return time_diff.days

    @property
    def sample_map(self):
        """Return dict of samples."""
        return {sample.name: sample for sample in self.samples}

    @property
    def analysis_date(self):
        """Date of the analysis run."""
        return self.analyzed_at.date()

    def to_dict(self, skip_samples=False):
        """Also include samples in the dict serialization."""
        analysis_dict = super(AnalysisRun, self).to_dict()
        analysis_dict['name'] = self.case.name
        if not skip_samples:
            sample_dicts = [sample.to_dict() for sample in self.samples]
            analysis_dict['samples'] = sample_dicts
        return analysis_dict


class Sample(Model):
    """Sample record."""

    __table_args__ = (UniqueConstraint('name', 'run_id',
                                       name='_uc_name_run_id'),)

    id = Column(types.Integer, primary_key=True)
    name = Column(types.String(32))

    run_id = Column(types.Integer, ForeignKey('analysis_run.id'),
                    nullable=False)
    assets = orm.relationship('Asset', backref='sample')


class Asset(Model):
    """Asset/file belonging to an analysis."""

    id = Column(types.Integer, primary_key=True)
    original_path = Column(types.Text)
    path = Column(types.String(256), nullable=False, unique=True)
    category = Column(types.String(32))
    archive_type = Column(types.Enum(*ARCHIVE_TYPES))
    is_local = Column(types.Boolean, default=True)
    checksum = Column(types.String(128))

    run_id = Column(types.Integer, ForeignKey('analysis_run.id'),
                    nullable=False)
    sample_id = Column(types.Integer, ForeignKey('sample.id'))

    def basename(self):
        return path(self.original_path).basename()
